#include "fmod.bi"
#include "fbgfx.bi"
#include "GL/gl.bi"
#include "GL/glu.bi"
#include "createtex.bi"
using FB

enum SPRITE
    logo
    star
    spacebar
    number
    level
    gameover
    paused
    sammydead
    tammydead
    sammy
    tammy
    walltex
    snaketex
    foodtex
    yes
    no
    k1
    k2
    k3
    k4
    numplayers
    selectspeed
    accelerate
    count
end enum

enum COLOUR
    null
    wall
    sammy
    tammy
    num1
    num2
    num3
    num4
    num5
    num6
    num7
    num8
    num9
    numx
    count
end enum

type snaketype
    clength   as integer
    length    as integer
    row       as integer
    col       as integer
    direction as integer
    lives     as integer
    score     as integer
    alive     as integer
end type

type colortype
    r as double
    g as double
    b as double
end type

type snakepoint
    x as integer
    y as integer
end type

const XRES = 800
const YRES = 600

const XMIN = -1.0*(XRES/YRES)
const YMIN = -1.0
const XMAX =  1.0*(XRES/YRES)
const YMAX =  1.0
const GFX_FLAGS = GFX_WINDOWED

const zoomTime = 1.0
const PI = 3.141592

const TRUE = 1
const FALSE = 0

const MAX_SNAKE_LENGTH = 1000

const SPEED_SLOW = 98
const SPEED_NORMAL = 107
const SPEED_FAST = 110
const SPEED_INSANE = 112

const MAX_LIVES = 5

#define SPEED_DELAY (adjustSpeed((120-speed)/200))
#define RAND (fix(rnd*1000000))

#define MUSIC_INTRO "music/nibbles.mp3"
#define MUSIC_GO "music/nibbles_go.mp3"
#define MUSIC_DIE "music/nibbles_die.mp3"
#define MUSIC_EAT "music/nibbles_eat.mp3"
#define MUSIC_WIN "music/nibbles_win.mp3"

declare sub init ( )
declare sub clsScr ( )

declare sub zoomOut ( byval sprite as integer, byval ix as double, byval iy as double )
declare sub zoomIn ( byval sprite as integer, byval ix as double, byval iy as double )

declare sub drawSprite ( byval spriteN as integer, byval x as double, byval y as double, byval a as double, byval zh as double, byval zv as double, byval r as double, byval g as double, byval b as double, byval alpha as double )
declare sub drawNumber ( byval num as integer, byval x as double, byval y as double, byval a as double, byval zh as double, byval zv as double, byval r as double, byval g as double, byval b as double, byval alpha as double )

declare sub drawWalls ( byval alpha as double, byval elev as double, byval angle as double, byval zoom as double, byval stAlpha as double = 1.0 )
declare sub drawStatus ( byval snake as integer, byval w as double, byval h as double, byval xc as double, byval yc as double, byval alpha as double = 1.0 )
declare sub drawStats ( byval alpha as double )

declare sub intro ( )
declare sub sparklePause ( )

declare sub __end ()

declare sub genLevel ( byval level as integer )

declare function yesNo ( byval image as integer, byval scale as double, byval tmax as double, byval levelAlpha as double, byval mode as integer = 0 ) as integer
declare function choice ( byval image as integer, byval scale as double, byval tmax as double, byval levelAlpha as double, byval options as integer, byval mode as integer = 0 ) as integer

declare function playLevel ( byval level as integer ) as integer

declare function adjustSpeed ( byval curDelay as double ) as double

dim shared as texture_t spr(SPRITE.count)
dim shared as texture_t levelSpr

dim shared as snaketype sammy(1 to 2)

dim shared as integer wall(1 to 50, 1 to 80)
dim shared as integer dirs(1 to 50, 1 to 80), pdirs(1 to 50, 1 to 80)

dim shared as colortype colorTable(COLOUR.count)
dim shared as snakepoint spoints(1 to 2, MAX_SNAKE_LENGTH)

dim shared as integer NumPlayers
dim shared as integer speed

dim shared as integer soundOn = TRUE
dim shared as FSOUND_STREAM ptr stream

dim shared as integer accelerate = TRUE
dim shared as integer speedInc = 0

' main : ------------------

init()
intro()

do
    NumPlayers = 0
    while NumPlayers = 0
        NumPlayers = choice(SPRITE.numplayers, 2.0, 1.0, 0.0, 2)
        if NumPlayers = 0 then
            __end()
        end if
    wend
    dim as integer sp = choice(SPRITE.selectspeed, 2.0, 1.0, 0.0, 4)
    if sp = 0 then
        __end()
    elseif sp = 1 then
        speed = SPEED_SLOW
    elseif sp = 2 then
        speed = SPEED_NORMAL
    elseif sp = 3 then
        speed = SPEED_FAST
    elseif sp = 4 then
        speed = SPEED_INSANE
    end if
    accelerate = yesNo(SPRITE.accelerate, 2.0, 1.0, 0.0)
    exit do
loop
dim as integer startLevel = 1

dim as integer level = startLevel
while 1
    if level = startLevel then
        dim as integer i, j
        for i = 1 to 2
            sammy(i).lives = MAX_LIVES
            sammy(i).score = 0
        next i
        speedInc = 0
        if NumPlayers = 1 then sammy(2).lives = 0
    end if
    level = playLevel(level)
    if level > 10 then
        level = 10
    elseif level = 0 then
        if yesNo(SPRITE.gameOver, 2.0, 1.0, 0.0) then
            level = startLevel
        else
            __end()
        end if
    end if
wend

__end()

' __end : -----------------

sub __end ()
    FSOUND_Close 
    end
end sub

' playMp3 : ---------------

sub playMp3 ( filename as string )

    if soundOn = TRUE then
        if stream <> 0 then
            FSOUND_Stream_Stop stream
            FSOUND_Stream_Close stream
        end if
    
        stream = FSOUND_Stream_Open( filename, FSOUND_MPEGACCURATE, 0, 0 ) 
        FSOUND_Stream_Play( FSOUND_FREE, stream )
    end if
    
end sub

' adjustSpeed : -----------

function adjustSpeed ( curDelay as double ) as double
    
    dim as double factor = 0.815 + 0.185 - (speedInc/8.0) * 0.185
    return curDelay * factor
    
end function

' init : ------------------

sub init ( )

    if FSOUND_Init( 44100, 4, 0 ) = 0 then
        soundOn = FALSE
    end if
    
    screenres XRES, YRES, 32, 2, GFX_FLAGS or GFX_OPENGL or GFX_MULTISAMPLE
    setmouse , , 0

    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(XMIN, XMAX, YMAX, YMIN, -1, 1)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    glClearColor(0.0, 0.0, 0.0, 1.0)
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST)
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST)
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE)
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE)
    glDisable(GL_DEPTH_TEST)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    glEnable(GL_BLEND)
    glEnable(GL_TEXTURE_2D)
    glEnable(GL_LINE_SMOOTH)
    
    #define LOADING clsScr(): glLineWidth(10.0): glBegin(GL_LINES): if alt = 0 then: glColor4f(1.0, 1.0, 1.0, 1.0): else: glColor4f(0.5, 0.5, 0.5, 1.0): end if : glVertex3f(XMIN, 0, 0): glVertex3f(XMAX, 0, 0): glEnd(): alt = (alt + 1) and 1: flip
    
    dim as integer alt = 0
    
    spr(SPRITE.logo) = loadTexture("gfx/logo.bmp", rgb(0, 0, 0)): LOADING
    spr(SPRITE.star) = loadTexture("gfx/star.bmp", rgb(255, 0, 255), TEX_MASKED): LOADING
    spr(SPRITE.spacebar) = loadTexture("gfx/space.bmp", rgb(0, 0, 0)): LOADING
    spr(SPRITE.number) = loadTexture("gfx/numbers.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.level) = loadTexture("gfx/level.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.gameover) = loadTexture("gfx/gameover.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.paused) = loadTexture("gfx/paused.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.sammydead) = loadTexture("gfx/sammydead.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.tammydead) = loadTexture("gfx/tammydead.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.sammy) = loadTexture("gfx/sammy.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.tammy) = loadTexture("gfx/tammy.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.walltex) = loadTexture("gfx/walltex.bmp", rgb(0, 0, 0)): LOADING
    spr(SPRITE.snaketex) = loadTexture("gfx/snaketex.bmp", rgb(0, 0, 0)): LOADING
    spr(SPRITE.foodtex) = loadTexture("gfx/foodtex.bmp", rgb(0, 0, 0)): LOADING
    spr(SPRITE.sammy) = loadTexture("gfx/sammy.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.yes) = loadTexture("gfx/yes.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.no) = loadTexture("gfx/no.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.numplayers) = loadTexture("gfx/numplayers.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING    
    spr(SPRITE.selectspeed) = loadTexture("gfx/selectspeed.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING    
    spr(SPRITE.k1) = loadTexture("gfx/1.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.k2) = loadTexture("gfx/2.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.k3) = loadTexture("gfx/3.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.k4) = loadTexture("gfx/4.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING
    spr(SPRITE.accelerate) = loadTexture("gfx/accelerate.bmp", rgb(0, 0, 0), TEX_ALPHAFROMRED or TEX_HASALPHA): LOADING

    with colorTable(COLOUR.null)
        .r = 0.0
        .g = 0.0
        .b = 0.0
    end with

    with colorTable(COLOUR.wall)
        .r = 1.0
        .g = 0.0
        .b = 0.0
    end with

    with colorTable(COLOUR.sammy)
        .r = 1.0
        .g = 1.0
        .b = 0.05
    end with

    with colorTable(COLOUR.tammy)
        .r = 1.0
        .g = 0.05
        .b = 1.0
    end with

end sub

' clsScr : ----------------

sub clsScr ( )
    
    glClear GL_COLOR_BUFFER_BIT
    glLoadIdentity()
    
end sub

' zoomOut : ---------------

sub zoomOut ( byval sprite as integer, byval ix as double, byval iy as double )
    
    dim as double st = TIMER
    dim as double t
    
    do
        t = (TIMER-st)/zoomTime
        if t > 1.0 then t = 1.0
        
        clsScr()
        glLoadIdentity()
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, spr(sprite).id)
        
        dim as double zoom = 1.0/(t+0.000001)
        
        glTranslatef(ix, iy, 0)
        glRotatef(-((1.0-t)*20)^2.25, 0, 0, 1.0)
        
        glColor4f(1.0, 1.0, 1.0, 1.0)
        glBegin(GL_QUADS)
            glTexCoord2f(0, 0): glVertex3f((XMIN-ix)*zoom, (YMIN-iy)*zoom, 0)
            glTexCoord2f(0, spr(sprite).vone): glVertex3f((XMIN-ix)*zoom, (YMAX-iy)*zoom, 0)
            glTexCoord2f(spr(sprite).hone, spr(sprite).vone): glVertex3f((XMAX-ix)*zoom, (YMAX-iy)*zoom, 0)
            glTexCoord2f(spr(sprite).hone, 0): glVertex3f((XMAX-ix)*zoom, (YMIN-iy)*zoom, 0)
        glEnd()
        
        flip
    loop until t >= 1.0
    
end sub

' zoomIn : ----------------

sub zoomIn ( byval sprite as integer, byval ix as double, byval iy as double )
    
    dim as double st = TIMER
    dim as double t
    
    do
        t = (TIMER-st)/(zoomTime*1.25)
        if t > 1.0 then t = 1.0
        
        clsScr()
        glLoadIdentity()
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, spr(sprite).id)
        
        dim as double zoom = ((1.0-t)+0.000001)^3
        
        glTranslatef(ix, iy, 0)
        'glRotatef(((t)*20)^2.25, 0, 0, 1.0)
        
        glColor4f(1.0, 1.0, 1.0, 1.0)
        glBegin(GL_QUADS)
            glTexCoord2f(0, 0): glVertex3f((XMIN-ix)*zoom, (YMIN-iy), 0)
            glTexCoord2f(0, spr(sprite).vone): glVertex3f((XMIN-ix)*zoom, (YMAX-iy), 0)
            glTexCoord2f(spr(sprite).hone, spr(sprite).vone): glVertex3f((XMAX-ix)*zoom, (YMAX-iy), 0)
            glTexCoord2f(spr(sprite).hone, 0): glVertex3f((XMAX-ix)*zoom, (YMIN-iy), 0)
        glEnd()
        
        flip
    loop until t >= 1.0
    
end sub

' drawNumber : ------------

sub drawNumber ( byval num as integer, byval x as double, byval y as double, byval a as double, byval zh as double, byval zv as double, byval r as double, byval g as double, byval b as double, byval alpha as double )
    dim as integer spriteN = SPRITE.number
    glPushMatrix()
    glBindTexture(GL_TEXTURE_2D, spr(spriteN).id)
    glEnable(GL_TEXTURE_2D)

    glLoadIdentity()
    glTranslatef(x, y, 0.0)
    glScaled(zh, zv, 0)
    glRotatef((a*180/PI) mod 360.0, 0, 0, 1.0)
    
    dim as double tx1, tx2
    
    tx1 = (spr(spriteN).hone / 11) * num
    tx2 = (spr(spriteN).hone / 11) * (num+1)
    
    glBegin(GL_QUADS)
      glColor4f(r, g, b, alpha)
      glTexCoord2f(tx1, 0): glVertex3f(-1, 1, 0)
      glTexCoord2f(tx2, 0): glVertex3f(1, 1, 0)
      glTexCoord2f(tx2, spr(spriteN).vone): glVertex3f(1, -1, 0)
      glTexCoord2f(tx1, spr(spriteN).vone): glVertex3f(-1, -1, 0)
    glEnd()
    glPopMatrix()
end sub

' drawSprite : ------------

sub drawSprite ( byval spriteN as integer, byval x as double, byval y as double, byval a as double, byval zh as double, byval zv as double, byval r as double, byval g as double, byval b as double, byval alpha as double )
    glPushMatrix()
    glBindTexture(GL_TEXTURE_2D, spr(spriteN).id)
    glEnable(GL_TEXTURE_2D)

    glLoadIdentity()
    glTranslatef(x, y, 0.0)
    glScaled(zh, zv, 0)
    glRotatef((a*180/PI) mod 360.0, 0, 0, 1.0)
    
    glBegin(GL_QUADS)
      glColor4f(r, g, b, alpha)
      glTexCoord2f(0, 0): glVertex3f(-1, 1, 0)
      glTexCoord2f(spr(spriteN).hone, 0): glVertex3f(1, 1, 0)
      glTexCoord2f(spr(spriteN).hone, spr(spriteN).vone): glVertex3f(1, -1, 0)
      glTexCoord2f(0, spr(spriteN).vone): glVertex3f(-1, -1, 0)
    glEnd()
    glPopMatrix()
end sub

' intro : -----------------

sub intro ( )
    
    zoomOut(SPRITE.logo, 0.039999, -0.646875)
    playMp3(MUSIC_INTRO)
    sparklePause()
    zoomIn(SPRITE.logo, 0, 0)
    
end sub

' sparklePause : ----------

sub RectFunc ( byval t as double, byref x as double, byref y as double )
    
    t = (t*1000 mod 1000)/1000
    dim as double it' = ((t*1000) mod 250) / 250
    
    dim as double sz = XRES/YRES*2+2
    dim as double w = (XRES/YRES)/sz
    dim as double h = 1/sz
    
    if t < w then
        it = t*(1/w)
        x = it * (XMAX-XMIN) + XMIN
        y = YMIN
    elseif t >= w and t < (w+h) then
        it = (t-w)*(1/h)
        x = XMAX
        y = it * (YMAX-YMIN) + YMIN
    elseif t >= (w+h) and t < (w+h+w) then
        it = (t-(w+h))*(1/w)
        x = (1.0-it) * (XMAX-XMIN) + XMIN
        y = YMAX
    elseif t >= (w+h+w) then
        it = (t-(w+h+w))*(1/h)
        x = XMIN
        y = (1.0-it) * (YMAX-YMIN) + YMIN
    end if
    
end sub

sub sparklePause ( )
    
    while multikey(SC_SPACE): flip: wend
    
    dim as double refTime = TIMER
    dim as double flipTime = rnd*5.0+5.0
    dim as double sign = 1
    
    dim as double startTime = TIMER
    dim as double toff = 0
    
    dim as double endTime = 0
    dim as double startT = 0
    
    while 1

        clsScr()
        glLoadIdentity()
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, spr(SPRITE.logo).id)
                        
        glColor4f(1.0, 1.0, 1.0, 1.0)
        glBegin(GL_QUADS)
            glTexCoord2f(0, 0): glVertex3f(XMIN, YMIN, 0)
            glTexCoord2f(0, spr(SPRITE.logo).vone): glVertex3f(XMIN, YMAX, 0)
            glTexCoord2f(spr(SPRITE.logo).hone, spr(SPRITE.logo).vone): glVertex3f(XMAX, YMAX, 0)
            glTexCoord2f(spr(SPRITE.logo).hone, 0): glVertex3f(XMAX, YMIN, 0)
        glEnd()
                
        dim as double x, y
        dim as double t
        dim as double et
        
        if endTime = 0 then
            et = 0
        else
            et = (TIMER - endTime)*2.0
        end if

        dim as double TIMERx = TIMER+toff

        if endTime > 0 then TIMERx = endTime

        dim as double tmr = TIMERx - refTime
        if tmr > flipTime and endTime <= 0 then
            if sign = -1 then toff -= tmr
            TIMERx = TIMER+toff
            flipTime = rnd*5.0+5.0
            sign = -sign
            refTime = TIMERx
            tmr = TIMERx - refTime
        end if
        tmr = tmr*sign + refTime

        if multikey(SC_SPACE) and endTime = 0 then
            dim as double t, x, y
            dim as double mint, mind = 100
            for t = 0 to 1.0 step (1.0/56)
                RectFunc(int((t+(tmr/3))*56)/56, x, y)
                if (y+abs(x)) < mind then
                    mind = (abs(x)+y)
                    mint = t
                end if
            next t
            startT = mint
            endTime = TIMER
        end if

        for t = (fix(et*55)/56.0)+startT to 1.0+startT step (1.0/56)
            RectFunc(int((t+(tmr/3))*56)/56, x, y)
            x *= 0.955: y *= 0.94
            dim as double sz = (sin(tmr+t*100)+1.0)/4.0 + 0.5
            dim as double tz = (TIMER - startTime)*3.0
            if tz > 1.0 then tz = 1.0
            drawSprite(SPRITE.star, x, y, tmr, (0.025*sz*1.5 + 0.025*0.5)*1, (0.025*sz*1.5 + 0.025*0.5)*1, 1.0, 0.0, 0.0, (sz*0.75)*tz)
        next t
        
        drawSprite(SPRITE.spacebar, 0, 0.93, 0.0, 0.25, -0.25/3.0, 0.0, 0.0, 0.0, 1.0*(1.0-et))
        drawSprite(SPRITE.spacebar, 0, 0.93, 0.0, 0.25, -0.25/4.0, 0.0, 1.0, 0.0, ((sin(TIMER*5)+1.0)/4.0 + 0.5)*(1-et))
        
        if multikey(SC_ESCAPE) then __end()

        flip
        
        if et > 1 then exit while
        
    wend
    
end sub

' initLevel : -------------

sub initLevel ( byval level as integer )
    
    dim as integer i, j

    sammy(1).clength = 1
    sammy(1).length = 2
    sammy(1).alive = TRUE
    sammy(2).clength = 1
    sammy(2).length = 2
    sammy(2).alive = TRUE

    for i = 1 to 80
        for j = 1 to 2
            wall(j,i) = 0
        next j
        for j = 3 to 50
            if i > 1 and i < 80 and j > 3 and j < 50 then
                wall(j,i) = 0
            else
                wall(j,i) = 1
            end if
        next j
    next i
    
    select case level
    case 1
        sammy(1).row = 25: sammy(2).row = 25
        sammy(1).col = 50: sammy(2).col = 30
        sammy(1).direction = 4: sammy(2).direction = 3


    case 2
        for i = 20 to 60
            wall(25, i) = 1
        next i
        sammy(1).row = 7: sammy(2).row = 43
        sammy(1).col = 60: sammy(2).col = 20
        sammy(1).direction = 3: sammy(2).direction = 4

    case 3
        for i = 10 to 40
            wall(i, 20) = 1
            wall(i, 60) = 1
        next i
        sammy(1).row = 25: sammy(2).row = 25
        sammy(1).col = 50: sammy(2).col = 30
        sammy(1).direction = 1: sammy(2).direction = 2

    case 4
        for i = 4 to 30
            wall(i, 20) = 1
            wall(53 - i, 60) = 1
        next i
        for i = 2 to 40
            wall(38, i) = 1
            wall(15, 81 - i) = 1
        next i
        sammy(1).row = 7: sammy(2).row = 43
        sammy(1).col = 60: sammy(2).col = 20
        sammy(1).direction = 3: sammy(2).direction = 4
   
    case 5
        for i = 13 to 39
            wall(i, 21) = 1
            wall(i, 59) = 1
        next i
        for i = 23 to 57
            wall(11, i) = 1
            wall(41, i) = 1
        next i
        sammy(1).row = 25: sammy(2).row = 25
        sammy(1).col = 50: sammy(2).col = 30
        sammy(1).direction = 1: sammy(2).direction = 2

    case 6
        for i = 4 to 49
            if i > 30 or i < 23 then
                wall(i, 10) = 1
                wall(i, 20) = 1
                wall(i, 30) = 1
                wall(i, 40) = 1
                wall(i, 50) = 1
                wall(i, 60) = 1
                wall(i, 70) = 1
            end if
        next i
        sammy(1).row = 7: sammy(2).row = 43
        sammy(1).col = 65: sammy(2).col = 15
        sammy(1).direction = 2: sammy(2).direction = 1

    case 7
        for i = 4 to 49 step 2
            wall(i, 40) = 1
        next i
        sammy(1).row = 7: sammy(2).row = 43
        sammy(1).col = 65: sammy(2).col = 15
        sammy(1).direction = 2: sammy(2).direction = 1

    case 8
        for i = 4 to 40
            wall(i, 10) = 1
            wall(53 - i, 20) = 1
            wall(i, 30) = 1
            wall(53 - i, 40) = 1
            wall(i, 50) = 1
            wall(53 - i, 60) = 1
            wall(i, 70) = 1
        next i
        sammy(1).row = 7: sammy(2).row = 43
        sammy(1).col = 65: sammy(2).col = 15
        sammy(1).direction = 2: sammy(2).direction = 1

    case 9
        for i = 6 to 47
            wall(i, i) = 1
            wall(i, i + 28) = 1
        next i
        sammy(1).row = 40: sammy(2).row = 15
        sammy(1).col = 75: sammy(2).col = 5
        sammy(1).direction = 1: sammy(2).direction = 2
   
    case else
        for i = 4 to 49 step 2
            wall(i, 10) = 1
            wall(i + 1, 20) = 1
            wall(i, 30) = 1
            wall(i + 1, 40) = 1
            wall(i, 50) = 1
            wall(i + 1, 60) = 1
            wall(i, 70) = 1
        next i
        sammy(1).row = 7: sammy(2).row = 43
        sammy(1).col = 65: sammy(2).col = 15
        sammy(1).direction = 2: sammy(2).direction = 1

    end select
    
    for i = 1 to NumPlayers
        dim as integer r = sammy(i).row, c = sammy(i).col
        select case sammy(i).direction
            case 1: r -= 1
            case 2: r += 1
            case 3: c -= 1
            case 4: c += 1
        end select
        wall(r, c) = COLOUR.sammy + i - 1
        dirs(r, c) = sammy(i).direction
        pdirs(r, c) = sammy(i).direction
        select case sammy(i).direction
            case 1: r -= 1
            case 2: r += 1
            case 3: c -= 1
            case 4: c += 1
        end select
        wall(r, c) = COLOUR.sammy + i - 1
        dirs(r, c) = sammy(i).direction
        pdirs(r, c) = sammy(i).direction
    next i

    'if levelSpr.id > 0 then
    '    glDeleteTextures(1, @levelSpr.id)
    'end if
   ' 
   ' dim as any ptr bfr = imagecreate(800, 600)
   ' 
   ' levelSpr = makeTexture(bfr, rgb(0, 0, 0))
    
end sub

' playLevel : -------------

sub levelIntro ( )
    
    dim as double st, t
    
    st = timer
    do
        t = (timer - st) * 1.0
        if t > 1.0 then exit do
        clsScr()
        drawWalls(t, t, t, t)
        drawStats(t)
        flip
    loop

end sub

sub levelSpacePause ( byval level as integer, byval ttime as double = 1.0 )

    dim as double st, t, t2, set = 0
    st = timer
    do
        t = (timer - st) / ttime
        if t > 1.0 then t = 1.0
        t2 = (timer - (st+0.5)) / ttime
        if t > 1.0 then t = 1.0
        if t2 > 1.0 then t2 = 1.0
        clsScr()
        drawWalls(1, 1, 1, 1)
        drawStats(1)
        glPushMatrix()
            glDisable(GL_TEXTURE_2D)
            glBegin(GL_QUADS)
                glColor4f(0,0,0,t*0.5)
                glVertex3f(XMIN, YMIN, 0)
                glVertex3f(XMAX, YMIN, 0)
                glVertex3f(XMAX, YMAX, 0)
                glVertex3f(XMIN, YMAX, 0)
            glEnd()
        glPopMatrix()
        drawSprite(SPRITE.level, 0.06 + -(XMAX-XMIN)*0.125*0.9*0.5, 0, 0, (XMAX-XMIN)*0.125*t, -(YMAX-YMIN)*0.125*0.5*t, 1.0, 1.0, 1.0, 1.0)
        if t2 > 0 then
            drawNumber(level, 0.06 + (XMAX-XMIN)*0.125*t,0, 0, (XMAX-XMIN)*0.125*0.5*0.9, -(YMAX-YMIN)*0.125*0.5*0.9, 1.0, 0.0, 0.0, t2)
        end if
        if t2 >= 0.5 then
            if set = 0 then set = timer
            dim as double et = (timer - set) * 4.0
            if et > 1.0 then et = 1.0
            'drawSprite(SPRITE.spacebar, 0, 0.25, 0.0, 0.25, -0.25/3.0, 0.0, 0.0, 0.0, 1.0*(et))
            drawSprite(SPRITE.spacebar, 0, 0.25, 0.0, 0.25, -0.25/4.0, 0.0, 1.0, 0.0, ((sin(TIMER*5)+1.0)/4.0 + 0.5)*(et))
        end if
        flip
        if multikey(SC_ESCAPE) then __end()
    loop until t2 >= 1.0 and multikey(SC_SPACE)
    
end sub

function yesNo ( byval image as integer, byval scale as double, byval tmax as double, byval levelAlpha as double, byval mode as integer = 0 ) as integer

    dim as double st, t
    st = timer
    do
        t = (timer - st)/tmax
        if t > 1.0 then t = 1.0
        clsScr()
        if image = SPRITE.sammydead or image = SPRITE.tammydead then
        elseif image = SPRITE.gameover then
            drawStats(t)
        else
            drawWalls(levelAlpha, 1, 1, 1)
            drawStats(levelAlpha)
        end if
        if mode = 0 then
            drawSprite(image, 0, 0, 0, (XMAX-XMIN)*0.125*t*scale, -(YMAX-YMIN)*0.125*0.5*t, 1.0, 1.0, 1.0, 1.0)
        else
            drawSprite(image, 0, 0, 0, (XMAX-XMIN)*0.125*scale, -(YMAX-YMIN)*0.125*0.5, 1.0, 1.0, 1.0, 1.0*t*t)
        end if
        dim as double et = ((timer) - (st+tmax/2)) * 4/(tmax/2)
        if et > 1.0 then et = 1.0
        drawSprite(SPRITE.spacebar, 0, 0.25, 0.0, 0.25, -0.25/3.0, 0.0, 0.0, 0.0, 1.0*(et))
        drawSprite(SPRITE.yes, 0-0.25/2.0, 0.25, 0.0, 0.25/3.0, -0.25/3.0, 0.0, 1.0, 0.0, ((sin(TIMER*5)+1.0)/4.0 + 0.5)*(et))
        drawSprite(SPRITE.no, 0+0.25/2.0, 0.25, 0.0, 0.25/3.0, -0.25/3.0, 1.0, 0.0, 0.0, ((sin(TIMER*5)+1.0)/4.0 + 0.5)*(et))
        
        if multikey(SC_ESCAPE) then __end()
        
        flip
        if et >= 1.0 then
            if multikey(SC_Y) then return TRUE
            if multikey(SC_N) then return FALSE
        end if
    loop
    
end function

function choice ( byval image as integer, byval scale as double, byval tmax as double, byval levelAlpha as double, byval options as integer, byval mode as integer = 0 ) as integer

    dim as integer ret = 0
    dim as double st, t
    st = timer
    do
        t = (timer - st)/tmax
        if t > 1.0 then t = 1.0
        clsScr()
        if image = SPRITE.sammydead or image = SPRITE.tammydead then
        elseif image = SPRITE.gameover then
            drawStats(t)
        else
            drawWalls(levelAlpha, 1, 1, 1)
            drawStats(levelAlpha)
        end if
        if mode = 0 then
            drawSprite(image, 0, 0, 0, (XMAX-XMIN)*0.125*t*scale, -(YMAX-YMIN)*0.125*0.5*t, 1.0, 1.0, 1.0, 1.0)
        else
            drawSprite(image, 0, 0, 0, (XMAX-XMIN)*0.125*scale, -(YMAX-YMIN)*0.125*0.5, 1.0, 1.0, 1.0, 1.0*t*t)
        end if
        dim as double et = ((timer) - (st+tmax/2)) * 4/(tmax/2)
        if et > 1.0 then et = 1.0
        drawSprite(SPRITE.spacebar, 0, 0.25, 0.0, 0.25, -0.25/3.0, 0.0, 0.0, 0.0, 1.0*(et))
        dim as integer i
        dim as double x1, x2
        x1 = 0.25/6.0 * -(options)*1.5
        x2 = 0.25/6.0 * +(options)*1.5
        for i = 1 to options
            dim as double tt = (i-1)/(options-1)
            drawSprite(SPRITE.k1 + i-1, (x2-x1)*tt+x1, 0.25, 0.0, 0.25/3.0, -0.25/3.0, tt, 1.0-tt, 0.0, et*0.5 + et*sin((TIMER+tt)*2)*0.25)
        next i
               
        flip
        if et >= 1.0 then
            dim as string k = inkey
            if k = chr(27) then: ret = 0: exit do: end if
            if k = "1" and options >= 1 then: ret = 1: exit do: end if
            if k = "2" and options >= 2 then: ret = 2: exit do: end if
            if k = "3" and options >= 3 then: ret = 3: exit do: end if
            if k = "4" and options >= 4 then: ret = 4: exit do: end if
        end if
    loop
    
    while len(inkey): wend
    
    return ret
    
end function

sub spacePause ( byval image as integer, byval scale as double, byval tmax as double, byval mode as integer = 0 )

    dim as double st, t
    st = timer
    do
        t = (timer - st)/tmax
        if t > 1.0 then t = 1.0
        clsScr()
        if image = SPRITE.sammydead or image = SPRITE.tammydead then
            drawStats(t)
        else
            drawWalls(0.5, 1, 1, 1)
            drawStats(0.5)
        end if
        if mode = 0 then
            drawSprite(image, 0, 0, 0, (XMAX-XMIN)*0.125*t*scale, -(YMAX-YMIN)*0.125*0.5*t, 1.0, 1.0, 1.0, 1.0)
        else
            drawSprite(image, 0, 0, 0, (XMAX-XMIN)*0.125*scale, -(YMAX-YMIN)*0.125*0.5, 1.0, 1.0, 1.0, 1.0*t*t)
        end if
        dim as double et = ((timer) - (st+tmax/2)) * 4/(tmax/2)
        if et > 1.0 then et = 1.0
        'drawSprite(SPRITE.spacebar, 0, 0.25, 0.0, 0.25, -0.25/3.0, 0.0, 0.0, 0.0, 1.0*(et))
        drawSprite(SPRITE.spacebar, 0, 0.25, 0.0, 0.25, -0.25/4.0, 0.0, 1.0, 0.0, ((sin(TIMER*5)+1.0)/4.0 + 0.5)*(et))
        flip
        if et >= 1.0 and multikey(SC_SPACE) then exit sub
        if multikey(SC_ESCAPE) then __end()
    loop
    
end sub

function playLevel ( byval level as integer ) as integer
    
    randomize TIMER
    
    initLevel(level)

    dim as integer i, j

    'levelIntro()
        
    dim as integer try = 0
        
    do
        initLevel(level)
        levelIntro()
        if try = 0 then
            playMp3(MUSIC_GO)
            levelSpacePause(level)
        else
            levelSpacePause(level, 0.25)
        end if
        try += 1
        dim as integer num = 0, curnum = 1
        speedInc = 0
        
        do
            if num <> curnum then
                dim as integer row, col
                row = 4 + (RAND mod 47): col = 1 + (RAND mod 80)
                while wall(row, col)
                    row = 4 + (RAND mod 47): col = 1 + (RAND mod 80)
                wend
                wall(row, col) = COLOUR.num1 + curnum - 1
                num = curnum
                if accelerate = TRUE then
                    speedInc = (num-1)
                else
                    speedInc = 0
                end if
            end if
            
            dim as double frameStart = TIMER
            
            dim as string kbd = inkey
            select case kbd
                case "w", "W": if sammy(2).direction <> 2 and sammy(2).clength > 2 and wall(sammy(2).row-1,sammy(2).col) = 0 then sammy(2).direction = 1
                case "s", "S": if sammy(2).direction <> 1 and sammy(2).clength > 2 and wall(sammy(2).row+1,sammy(2).col) = 0 then sammy(2).direction = 2
                case "a", "A": if sammy(2).direction <> 4 and sammy(2).clength > 2 and wall(sammy(2).row,sammy(2).col-1) = 0 then sammy(2).direction = 3
                case "d", "D": if sammy(2).direction <> 3 and sammy(2).clength > 2 and wall(sammy(2).row,sammy(2).col+1) = 0 then sammy(2).direction = 4
                case chr$(255) + "H": if sammy(1).direction <> 2 and sammy(1).clength > 2 and wall(sammy(1).row-1,sammy(1).col) = 0 then sammy(1).direction = 1
                case chr$(255) + "P": if sammy(1).direction <> 1 and sammy(1).clength > 2 and wall(sammy(1).row+1,sammy(1).col) = 0 then sammy(1).direction = 2
                case chr$(255) + "K": if sammy(1).direction <> 4 and sammy(1).clength > 2 and wall(sammy(1).row,sammy(1).col-1) = 0 then sammy(1).direction = 3
                case chr$(255) + "M": if sammy(1).direction <> 3 and sammy(1).clength > 2 and wall(sammy(1).row,sammy(1).col+1) = 0 then sammy(1).direction = 4
                case "p", "P": spacePause(SPRITE.paused, 1.0, 0.5)
                case else
            end select
                
            for i = 1 to NumPlayers
                dim as integer oldrow = sammy(i).row, oldcol = sammy(i).col
                select case sammy(i).direction
                    case 1: sammy(i).row -= 1
                    case 2: sammy(i).row += 1
                    case 3: sammy(i).col -= 1
                    case 4: sammy(i).col += 1
                end select
                
                if wall(sammy(i).row, sammy(i).col) = (COLOUR.num1 + curnum - 1) then
                    sammy(i).length += curnum * 4
                    sammy(i).score += curnum
                    curnum += 1
                    if curnum < 10 then
                        playMp3(MUSIC_EAT)
                    else
                        wall(sammy(i).row, sammy(i).col) = COLOUR.sammy + i - 1
                        dim as double startTime = TIMER
                        playMp3(MUSIC_WIN)
                        while (TIMER - startTime) < 0.5
                            clsScr()
                            dim as double t = (TIMER-startTime) / 0.5
                            drawWalls(1.0, 1.0-t, 1.0-t, 1.0-t, 1.0-t)
                            flip
                        wend
                        return level+1
                    end if
                elseif wall(sammy(i).row, sammy(i).col) > 0 and sammy(i).clength > 2 then
                    ' player I dies
                    sammy(i).lives -= 1
                    sammy(i).score *= 0.8
                    if sammy(i).score < 0 then sammy(i).score = 0
                    dim as double startTime = TIMER
                    while 1
                        dim as double tt = (TIMER - startTime) / 0.25
                        if tt > 1.0 then tt = 1.0
                        clsScr()
                        drawWalls(1, 1, 1, 1)
                        drawStats(1.0)
                        glLoadIdentity()
                        glDisable(GL_TEXTURE_2D)
                        glBegin(GL_QUADS)
                            glColor4f(1,1,1,tt)
                            glVertex3f(XMIN, YMIN, 0)
                            glVertex3f(XMAX, YMIN, 0)
                            glVertex3f(XMAX, YMAX, 0)
                            glVertex3f(XMIN, YMAX, 0)
                        glEnd()
                        flip
                        if tt = 1.0 then exit while 
                    wend
                    startTime = TIMER
                    while (TIMER - startTime) < 0.25: wend
                    playMp3(MUSIC_DIE)
                    startTime = TIMER
                    while (TIMER - startTime) < 1.25
                        clsScr()
                        'drawWalls(1.0 - (TIMER-startTime)/0.25, 1.0, 1.0, 1.0, 1.0)
                        'drawStats(1.0 - (TIMER-startTime)/0.25)
                        glLoadIdentity()
                        glDisable(GL_TEXTURE_2D)
                        glBegin(GL_QUADS)
                            glColor4f(1,1,1,1.0 - (TIMER-startTime)/1.25)
                            glVertex3f(XMIN, YMIN, 0)
                            glVertex3f(XMAX, YMIN, 0)
                            glVertex3f(XMAX, YMAX, 0)
                            glVertex3f(XMIN, YMAX, 0)
                        glEnd()
                        flip
                    wend
                    initlevel(level)
                    if sammy(i).lives > 0 then
                        if i = 1 then
                            spacePause(SPRITE.sammydead, 1.5, 2.0, 1)
                        else
                            spacePause(SPRITE.tammydead, 1.5, 2.0, 1)
                        end if
                        exit do
                    else
                        return 0
                    end if
                end if
                
                for j = sammy(i).clength-1 to 1 step -1
                    'dim as integer r1 = spoints(i, j).y, c1 = spoints(i, j).x
                    'dim as integer r2 = spoints(i, j-1).y, c2 = spoints(i, j-1).x
                    'if r1 > 0 and c1 > 0 then dirs(r1, c1) = dirs(r2, c2)
                    spoints(i, j) = spoints(i, j-1)
                next j
                pdirs(oldrow, oldcol) = sammy(i).direction
                spoints(i, 0).y = sammy(i).row
                spoints(i, 0).x = sammy(i).col
                dirs(sammy(i).row, sammy(i).col) = sammy(i).direction
                pdirs(sammy(i).row, sammy(i).col) = 0
                if sammy(i).clength = (sammy(i).length+1) then
                    wall(spoints(i, sammy(i).clength-1).y, spoints(i, sammy(i).clength-1).x) = 0
                end if
                wall(sammy(i).row, sammy(i).col) = COLOUR.sammy+i-1
                sammy(i).clength += 1
                if sammy(i).clength > (sammy(i).length+1) then sammy(i).clength = sammy(i).length+1
            next i
            
            do 
                clsScr()
                drawWalls(1.0, 1.0, 1.0, 1.0)
                drawStats(1.0)
                flip
                if multikey(SC_ESCAPE) then __end()
            loop until not ((TIMER-frameStart) < SPEED_DELAY)
        loop
    loop until sammy(1).lives = 0 or (sammy(2).lives = 0 and NumPlayers > 1)
    
end function

' drawStatus : ------------

sub drawStats ( byval alpha as double )
    
    drawStatus(1, (XMAX-XMIN)*0.5*0.9, (YMAX-YMIN)*(1.0-0.85)*0.9, XMAX - (XMAX-XMIN)*0.5*0.95*0.5, YMIN+(YMAX-YMIN)*(1.0-0.85)*0.8*0.7, alpha)
    if NumPlayers > 1 then
        drawStatus(2, (XMAX-XMIN)*0.5*0.9, (YMAX-YMIN)*(1.0-0.85)*0.9, XMIN + (XMAX-XMIN)*0.5*0.95*0.5, YMIN+(YMAX-YMIN)*(1.0-0.85)*0.8*0.7, alpha)
    else
        drawStatus(2, (XMAX-XMIN)*0.5*0.9, (YMAX-YMIN)*(1.0-0.85)*0.9, XMIN + (XMAX-XMIN)*0.5*0.95*0.5, YMIN+(YMAX-YMIN)*(1.0-0.85)*0.8*0.7, alpha*0.25)
    end if
    
end sub

sub drawStatus ( byval snake as integer, byval w as double, byval h as double, byval xc as double, byval yc as double, byval alpha as double  )
    
    #define SCX(X) ((X)*w+xc-(w/2))
    #define SCY(Y) ((Y)*h+yc-(h/2))
    
    #define SCW(W) (SCX(W)-SCX(0))
    #define SCH(H) (SCY(H)-SCY(0))
        
    glPushMatrix()
    glLoadIdentity()
    
    glEnable(GL_TEXTURE_2D)
    glBindTexture(GL_TEXTURE_2D, spr(SPRITE.sammy+snake-1).id)
    glBegin(GL_QUADS)
        glColor4f(colorTable(COLOUR.sammy+snake-1).r, colorTable(COLOUR.sammy+snake-1).g, colorTable(COLOUR.sammy+snake-1).b, alpha)
        glTexCoord2f(0, 0): glVertex3f(SCX(0), SCY(0), 0)
        glTexCoord2f(spr(SPRITE.sammy+snake-1).hone, 0): glVertex3f(SCX(0.25), SCY(0), 0)
        glTexCoord2f(spr(SPRITE.sammy+snake-1).hone, spr(SPRITE.sammy+snake-1).vone): glVertex3f(SCX(0.25), SCY(0.75), 0)
        glTexCoord2f(0, spr(SPRITE.sammy+snake-1).vone): glVertex3f(SCX(0), SCY(0.75), 0)
    glEnd()
    
    dim as integer i
    
    for i = 1 to MAX_LIVES
        glBindTexture(GL_TEXTURE_2D, spr(SPRITE.star).id)
        glBegin(GL_QUADS)
            if sammy(snake).lives >= i then
                glColor4f(1.0, 1.0, 1.0, alpha)
            else
                glColor4f(1.0, 0.0, 0.0, alpha)
            end if
            dim as double x1 = (0.25/MAX_LIVES)*(i-1), x2 = (0.25/MAX_LIVES)*(i) - (0.25/MAX_LIVES)*0.25
            glTexCoord2f(0, 0): glVertex3f(SCX(x1), SCY(0.75), 0)
            glTexCoord2f(spr(SPRITE.star).hone, 0): glVertex3f(SCX(x2), SCY(0.75), 0)
            glTexCoord2f(spr(SPRITE.star).hone, spr(SPRITE.star).vone): glVertex3f(SCX(x2), SCY(1.0), 0)
            glTexCoord2f(0, spr(SPRITE.star).vone): glVertex3f(SCX(x1), SCY(1.0), 0)
        glEnd()
    next i
    
    dim as integer score = sammy(snake).score * 10000
    
    dim as integer f0 = 1
    for i = 1 to 9
        dim as integer digit = fix(score/(10^(9-i))) mod 10
        if digit > 0 then f0 = 0
            dim as double r, g, b
            r = 1.0
            g = 1.0 - ((9-i))/6
            b = g
        if not (digit > 0 or f0 = 0 or i = 9) then
            r /= 2
            g /= 2
            b /= 2
        end if
        drawNumber(digit, SCX(0.3 + (0.7/9)*(i-0.5)), SCY(0.5), 0, SCW(0.7/9.0)/2.0, -SCH(1.0)/2.0, r, g, b, alpha)
    next i
    
    glPopMatrix()
    
end sub

' drawWalls : -------------

sub drawWalls ( byval alpha as double, byval elev as double, byval angle as double, byval zoom as double, byval stAlpha as double = 1.0 )

    dim as integer i, j
    
    static as double __st = 0
    
    if __st = 0 then __st = TIMER
    
    dim as double cellsx, cellsy, xoff, yoff
    dim as double yeloff
    dim as double tm1 = (fix((TIMER-__st)/SPEED_DELAY*100) mod 100)/100
    
    cellsx = (XMAX-XMIN) / 80.0
    cellsy = ((YMAX-YMIN)*0.85) / 50.0
    xoff = 0 - (cellsx*80.0) / 2
    yoff = 0 - (cellsy*50.0) / 2 + (1.0-0.85)*(YMAX-YMIN)*0.5
    yeloff = (elev)*(cellsy/1.25)
    
    glPushMatrix()
    glLoadIdentity()
    glDisable(GL_TEXTURE_2D)
    dim as double ix, iy
    if NumPlayers = 1 then
        i = 1
    else
        if sammy(2).score > sammy(1).score then
            i = 2
        else
            i = 1
        end if
    end if
    ix = ((sammy(i).col)-0.5)*cellsx + xoff
    iy = ((sammy(i).row)-0.5)*cellsy + yoff
    glTranslatef(ix, iy, 0)
    glRotatef(-((1.0-angle)*20)^2.1, 0, 0, 1.0)
    zoom = 1.0/(zoom+0.000001)

    for j = 2 to 50
        for i = 1 to 80
            if wall(j, i) = 0 then
                glDisable(GL_TEXTURE_2D)
                glBindTexture(GL_TEXTURE_2D, spr(SPRITE.walltex).id)
                glBegin(GL_QUADS)
                    if (i+j) and 1 then
                        glColor4f(0.025*alpha, 0.025*alpha, 0.125*alpha, stAlpha)
                    else
                        glColor4f(0.0*alpha, 0.0*alpha, 0.0*alpha, stAlpha)
                    end if
                    dim as double x, y
                    x = (i-0.5)*cellsx + xoff
                    y = (j-0.5)*cellsy + yoff
                    glVertex3f((x-(cellsx/2)-ix)*zoom, (y-(cellsy/2)-iy)*zoom, 0)
                    glVertex3f((x+(cellsx/2)-ix)*zoom, (y-(cellsy/2)-iy)*zoom, 0)
                    glVertex3f((x+(cellsx/2)-ix)*zoom, (y+(cellsy/2)-iy)*zoom, 0)
                    glVertex3f((x-(cellsx/2)-ix)*zoom, (y+(cellsy/2)-iy)*zoom, 0)
                glEnd()
            elseif wall(j, i) > 0 then
                dim as double r, g, b
                r = colorTable(wall(j, i)).r
                g = colorTable(wall(j, i)).g
                b = colorTable(wall(j, i)).b
                dim as double x, y
                x = (i-0.5)*cellsx + xoff
                y = (j-0.5)*cellsy + yoff
                dim as integer num = 0
                if wall(j, i) >= COLOUR.num1 and wall(j, i) <= COLOUR.num9 then
                    'glPushMatrix()
                    'drawNumber(wall(j, i) - COLOUR.num1 + 1, x, y-yeloff/1.5, PI+sin(TIMER)*PI/8, -cellsx/1.75, cellsy/1.75, 1.0, 1.0, 1.0, 1.0)
                    'glPopMatrix()
                    r = 0.3
                    g = 1.0
                    b = 0.3
                    num = 1
                end if
                dim as double oalpha = 1.0
                dim as double _i = i , _j = j  
                if wall(j, i) = COLOUR.sammy or wall(j, i) = COLOUR.tammy or num = 1 then
                    oalpha *= stAlpha
                    glEnable(GL_TEXTURE_2D)
                    
                    dim as double txoff=0, tyoff=0
                    
                    dim as integer rp = 0
                    
                    if num then
                        glBindTexture(GL_TEXTURE_2D, spr(SPRITE.foodtex).id)
                    else
                        glBindTexture(GL_TEXTURE_2D, spr(SPRITE.snaketex).id)
                        select case dirs(j, i)
                        case 1: tyoff = -(tm1+0.0001)
                        case 2: tyoff = +(tm1+0.0001)
                        case 3: txoff = -(tm1+0.0001)
                        case 4: txoff = +(tm1+0.0001)
                        end select
                        select case pdirs(j, i)
                        case 1: tyoff = -(tm1+0.0001)
                        case 2: tyoff = +(tm1+0.0001)
                        case 3: txoff = -(tm1+0.0001)
                        case 4: txoff = +(tm1+0.0001)
                        end select
                        if dirs(j, i)>0 and pdirs(j, i)>0 then
                            if (dirs(j, i) = 1) and (pdirs(j, i) = 3) then: rp = 2 ' +/-
                            elseif (dirs(j, i) = 1) and (pdirs(j, i) = 4) then: rp = 1 ' -/-
                            elseif (dirs(j, i) = 2) and (pdirs(j, i) = 3) then: rp = 3 ' +/+
                            elseif (dirs(j, i) = 2) and (pdirs(j, i) = 4) then: rp = 4 ' -/+
                            elseif (pdirs(j, i) = 1) and (dirs(j, i) = 3) then: rp = 4
                            elseif (pdirs(j, i) = 1) and (dirs(j, i) = 4) then: rp = 3
                            elseif (pdirs(j, i) = 2) and (dirs(j, i) = 3) then: rp = 1
                            elseif (pdirs(j, i) = 2) and (dirs(j, i) = 4) then: rp = 2
                            end if
                        end if
                    end if
                    if rp > 0 then
                        glBegin(GL_TRIANGLES)
                    else
                        glBegin(GL_QUADS)
                    end if
                        glColor4f(r*alpha, g*alpha, b*alpha, oalpha)
                        dim as double y1=-1000, y2=-1000, tx, ty
                        if rp <> 1 then
                            glTexCoord2f(0+txoff, 0+tyoff)
                            tx = (x-(cellsx/2)-ix)*zoom: ty = (y-(cellsy/2)-yeloff-iy)*zoom
                            glVertex3f(tx, ty, 0)
                            y1 = ty
                        end if
                        if rp <> 2 then
                            glTexCoord2f(1+txoff, 0+tyoff)
                            tx = (x+(cellsx/2)-ix)*zoom: ty = (y-(cellsy/2)-yeloff-iy)*zoom
                            glVertex3f(tx, ty, 0)
                            y2 = ty
                        end if
                        if rp <> 3 then
                            glTexCoord2f(1+txoff, 1+tyoff)
                            tx = (x+(cellsx/2)-ix)*zoom: ty = (y+(cellsy/2)-yeloff-iy)*zoom
                            if ty > y2 then y2 = ty
                            glVertex3f(tx, ty, 0)
                        end if
                        if rp <> 4 then
                            glTexCoord2f(0+txoff, 1+tyoff)
                            tx = (x-(cellsx/2)-ix)*zoom: ty = (y+(cellsy/2)-yeloff-iy)*zoom
                            glVertex3f(tx, ty, 0)
                            if ty > y1 then y1 = ty
                        end if
                    glEnd()
                    glBegin(GL_QUADS)
                        glColor4f(r*alpha, g*alpha, b*alpha, oalpha)
                        glTexCoord2f(0+txoff, 0+tyoff): glVertex3f((x-(cellsx/2)-ix)*zoom, y1, 0)
                        glTexCoord2f(1+txoff, 0+tyoff): glVertex3f((x+(cellsx/2)-ix)*zoom, y2, 0)
                        glColor4f(r*0.25*alpha, g*0.25*alpha, b*0.25*alpha, oalpha)
                        glTexCoord2f(1+txoff, 1+tyoff): glVertex3f((x+(cellsx/2)-ix)*zoom, y2+yeloff*zoom, 0)
                        glTexCoord2f(0+txoff, 1+tyoff): glVertex3f((x-(cellsx/2)-ix)*zoom, y1+yeloff*zoom, 0)
                    glEnd()
                else
                    dim as double tyoff = (fix(TIMER*125) mod 1000) / 1000
                    glEnable(GL_TEXTURE_2D)
                    glBindTexture(GL_TEXTURE_2D, spr(SPRITE.walltex).id)
                    glBegin(GL_QUADS)
                        glColor4f(r*alpha, g*alpha, b*alpha, oalpha)
                        glTexCoord2f(_i/20, _j/12.5+tyoff): glVertex3f((x-(cellsx/2)-ix)*zoom, (y-(cellsy/2)-yeloff-iy)*zoom, 0)
                        glTexCoord2f((_i+1)/20, _j/12.5+tyoff): glVertex3f((x+(cellsx/2)-ix)*zoom, (y-(cellsy/2)-yeloff-iy)*zoom, 0)
                        glTexCoord2f((_i+1)/20, (_j+1)/12.5+tyoff): glVertex3f((x+(cellsx/2)-ix)*zoom, (y+(cellsy/2)-yeloff-iy)*zoom, 0)
                        glTexCoord2f(_i/20, (_j+1)/12.5+tyoff): glVertex3f((x-(cellsx/2)-ix)*zoom, (y+(cellsy/2)-yeloff-iy)*zoom, 0)
                        glColor4f(r*alpha*0.75, g*alpha*0.75, b*alpha*0.75, oalpha*0.75)
                        glTexCoord2f(_i/20, (_j+1)/12.5+tyoff): glVertex3f((x-(cellsx/2)-ix)*zoom, (y+(cellsy/2)-yeloff-iy)*zoom, 0)
                        glTexCoord2f((_i+1)/20, (_j+1)/12.5+tyoff): glVertex3f((x+(cellsx/2)-ix)*zoom, (y+(cellsy/2)-yeloff-iy)*zoom, 0)
                        glColor4f(r*0.025*alpha, g*0.025*alpha, b*0.025*alpha, oalpha)
                        glTexCoord2f((_i+1)/20, (_j+2)/12.5+tyoff): glVertex3f((x+(cellsx/2)-ix)*zoom, (y+(cellsy/2)-iy)*zoom, 0)
                        glTexCoord2f(_i/20, (_j+2)/12.5+tyoff): glVertex3f((x-(cellsx/2)-ix)*zoom, (y+(cellsy/2)-iy)*zoom, 0)
                    glEnd()
                end if
                    
                if wall(j, i) >= COLOUR.num1 and wall(j, i) <= COLOUR.num9 then
                    glPushMatrix()
                    drawNumber(wall(j, i) - COLOUR.num1 + 1, x, y-yeloff/1.25, PI, -cellsx/1.125, cellsy/2, 0.0, 1.0, 0.0, 1.0*alpha)
                    glPopMatrix()
                end if
            end if
        next i
    next j
    glPopMatrix()

end sub
